@@ -113,24 +113,19 @@ module Sequel
       # is true.
       def apply_connection_settings
         if Postgres.force_standard_strings
-          sql = "SET standard_conforming_strings = ON"
-          @db.log_info(sql)
           # This setting will only work on PostgreSQL 8.2 or greater
           # and we don't know the server version at this point, so
           # try it unconditionally and rescue any errors.
-          execute(sql) rescue nil
+          execute("SET standard_conforming_strings = ON") rescue nil
         end
         if cmm = Postgres.client_min_messages
-          sql = "SET client_min_messages = '#{cmm.to_s.upcase}'"
-          @db.log_info(sql)
-          execute(sql)
+          execute("SET client_min_messages = '#{cmm.to_s.upcase}'")
         end
       end
 
       # Get the last inserted value for the given sequence.
       def last_insert_id(sequence)
         sql = SELECT_CURRVAL % sequence
-        @db.log_info(sql)
         execute(sql) do |r|
           val = single_value(r)
           return val.to_i if val
@@ -140,7 +135,6 @@ module Sequel
       # Get the primary key for the given table.
       def primary_key(schema, table)
         sql = SELECT_PK[schema, table]
-        @db.log_info(sql)
         execute(sql) do |r|
           return single_value(r)
         end
@@ -149,14 +143,12 @@ module Sequel
       # Get the primary key and sequence for the given table.
       def sequence(schema, table)
         sql = SELECT_SERIAL_SEQUENCE[schema, table]
-        @db.log_info(sql)
         execute(sql) do |r|
           seq = single_value(r)
           return seq if seq
         end
         
         sql = SELECT_CUSTOM_SEQUENCE[schema, table]
-        @db.log_info(sql)
         execute(sql) do |r|
           return single_value(r)
         end
@@ -165,10 +157,17 @@ module Sequel
     
     # Methods shared by Database instances that connect to PostgreSQL.
     module DatabaseMethods
+      EXCLUDE_SCHEMAS = %w[pg_catalog pg_toast pg_temp_1 pg_toast_temp_1 information_schema]
       PREPARED_ARG_PLACEHOLDER = LiteralString.new('$').freeze
       RE_CURRVAL_ERROR = /currval of sequence "(.*)" is not yet defined in this session|relation "(.*)" does not exist/.freeze
       SYSTEM_TABLE_REGEXP = /^pg|sql/.freeze
 
+      # Commit an existing prepared transaction with the given transaction
+      # identifier string.
+      def commit_prepared_transaction(transaction_id)
+        run("COMMIT PREPARED #{literal(transaction_id)}")
+      end
+
       # Creates the function in the database.  Arguments:
       # * name : name of the function to create
       # * definition : string definition of the function, or object file for a dynamically loaded C function.
@@ -261,14 +260,8 @@ module Sequel
         self << drop_trigger_sql(table, name, opts)
       end
       
-      # Return a hash containing index information. Hash keys are index name symbols.
-      # Values are subhashes with two keys, :columns and :unique.  The value of :columns
-      # is an array of symbols of column names.  The value of :unique is true or false
-      # depending on if the index is unique.
-      #
-      # This will not output any partial indexes, indexes that aren't valid or ready,
-      # or indexes that contain anything other than simple column references.
-      def indexes(table)
+      # Use the pg_* system tables to determine indexes on a table
+      def indexes(table, opts={})
         m = output_identifier_meth
         im = input_identifier_meth
         schema, table = schema_and_table(table)
@@ -332,10 +325,16 @@ module Sequel
         get{setval(seq, db[table].select{coalesce(max(pk)+seq_ds.select{:increment_by}, seq_ds.select(:min_value))}, false)}
       end
 
+      # Rollback an existing prepared transaction with the given transaction
+      # identifier string.
+      def rollback_prepared_transaction(transaction_id)
+        run("ROLLBACK PREPARED #{literal(transaction_id)}")
+      end
+
       # PostgreSQL uses SERIAL psuedo-type instead of AUTOINCREMENT for
       # managing incrementing primary keys.
       def serial_primary_key_options
-        {:primary_key => true, :type => :serial}
+        {:primary_key => true, :serial => true, :type=>Integer}
       end
       
       # The version of the PostgreSQL server, used for determining capability.
@@ -351,11 +350,23 @@ module Sequel
         @server_version
       end
       
+      # PostgreSQL supports prepared transactions (two-phase commit) if
+      # max_prepared_transactions is greater than 0.
+      def supports_prepared_transactions?
+        return @supports_prepared_transactions if defined?(@supports_prepared_transactions)
+        @supports_prepared_transactions = self['SHOW max_prepared_transactions'].get.to_i > 0
+      end
+
       # PostgreSQL supports savepoints
       def supports_savepoints?
         true
       end
 
+      # PostgreSQL supports transaction isolation levels
+      def supports_transaction_isolation_levels?
+        true
+      end
+
       # Whether the given table exists in the database
       #
       # Options:
@@ -376,13 +387,24 @@ module Sequel
       # * :schema - The schema to search (default_schema by default)
       # * :server - The server to use
       def tables(opts={})
-        ds = metadata_dataset.from(:pg_class).filter(:relkind=>'r').select(:relname).exclude(SQL::StringExpression.like(:relname, SYSTEM_TABLE_REGEXP)).server(opts[:server]).join(:pg_namespace, :oid=>:relnamespace, :nspname=>(opts[:schema]||default_schema||'public').to_s) 
+        ds = metadata_dataset.from(:pg_class).filter(:relkind=>'r').select(:relname).exclude(SQL::StringExpression.like(:relname, SYSTEM_TABLE_REGEXP)).server(opts[:server]).join(:pg_namespace, :oid=>:relnamespace) 
+        ds = filter_schema(ds, opts)
         m = output_identifier_meth
         block_given? ? yield(ds) : ds.map{|r| m.call(r[:relname])}
       end
 
       private
-      
+
+      # If the :prepare option is given and we aren't in a savepoint,
+      # prepare the transaction for a two-phase commit.
+      def commit_transaction(conn, opts={})
+        if opts[:prepare] && Thread.current[:sequel_transaction_depth] <= 1
+          log_connection_execute(conn, "PREPARE TRANSACTION #{literal(opts[:prepare])}")
+        else
+          super
+        end
+      end
+
       # SQL statement to create database function.
       def create_function_sql(name, definition, opts={})
         args = opts[:args]
@@ -441,6 +463,16 @@ module Sequel
         "DROP TRIGGER#{' IF EXISTS' if opts[:if_exists]} #{name} ON #{quote_schema_table(table)}#{' CASCADE' if opts[:cascade]}"
       end
 
+      # If opts includes a :schema option, or a default schema is used, restrict the dataset to
+      # that schema.  Otherwise, just exclude the default PostgreSQL schemas except for public.
+      def filter_schema(ds, opts)
+        if schema = opts[:schema] || default_schema
+          ds.filter(:pg_namespace__nspname=>schema.to_s)
+        else
+          ds.exclude(:pg_namespace__nspname=>EXCLUDE_SCHEMAS)
+        end
+      end
+      
       # PostgreSQL folds unquoted identifiers to lowercase, so it shouldn't need to upcase identifiers on input.
       def identifier_input_method_default
         nil
@@ -501,6 +533,11 @@ module Sequel
           nil
         end
       end
+
+      # Don't log, since logging is done by the underlying connection.
+      def log_connection_execute(conn, sql) 
+        conn.execute(sql)
+      end
       
       # Use a dollar sign instead of question mark for the argument
       # placeholder.
@@ -532,13 +569,14 @@ module Sequel
           from(:pg_class).
           join(:pg_attribute, :attrelid=>:oid).
           join(:pg_type, :oid=>:atttypid).
+          join(:pg_namespace, :oid=>:pg_class__relnamespace).
           left_outer_join(:pg_attrdef, :adrelid=>:pg_class__oid, :adnum=>:pg_attribute__attnum).
           left_outer_join(:pg_index, :indrelid=>:pg_class__oid, :indisprimary=>true).
           filter(:pg_attribute__attisdropped=>false).
           filter{|o| o.pg_attribute__attnum > 0}.
           filter(:pg_class__relname=>m2.call(table_name)).
           order(:pg_attribute__attnum)
-        ds.join!(:pg_namespace, :oid=>:pg_class__relnamespace, :nspname=>(opts[:schema] || default_schema).to_s) if opts[:schema] || default_schema
+        ds = filter_schema(ds, opts)
         ds.map do |row|
           row[:default] = nil if blank_object?(row[:default])
           row[:type] = schema_column_type(row[:db_type])
@@ -551,11 +589,21 @@ module Sequel
         "(#{Array(args).map{|a| Array(a).reverse.join(' ')}.join(', ')})"
       end
       
+      # Handle bigserial type if :serial option is present
+      def type_literal_generic_bignum(column)
+        column[:serial] ? :bigserial : super
+      end
+
       # PostgreSQL uses the bytea data type for blobs
       def type_literal_generic_file(column)
         :bytea
       end
 
+      # Handle serial type if :serial option is present
+      def type_literal_generic_integer(column)
+        column[:serial] ? :serial : super
+      end
+
       # PostgreSQL prefers the text datatype.  If a fixed size is requested,
       # the char type is used.  If the text type is specifically
       # disallowed or there is a size specified, use the varchar type.
@@ -578,11 +626,11 @@ module Sequel
       BOOL_FALSE = 'false'.freeze
       BOOL_TRUE = 'true'.freeze
       COMMA_SEPARATOR = ', '.freeze
+      DELETE_CLAUSE_METHODS = Dataset.clause_methods(:delete, %w'from using where')
       EXCLUSIVE = 'EXCLUSIVE'.freeze
       EXPLAIN = 'EXPLAIN '.freeze
       EXPLAIN_ANALYZE = 'EXPLAIN ANALYZE '.freeze
       FOR_SHARE = ' FOR SHARE'.freeze
-      FOR_UPDATE = ' FOR UPDATE'.freeze
       LOCK = 'LOCK TABLE %s IN %s MODE'.freeze
       NULL = LiteralString.new('NULL').freeze
       PG_TIMESTAMP_FORMAT = "TIMESTAMP '%Y-%m-%d %H:%M:%S".freeze
@@ -595,6 +643,7 @@ module Sequel
       SHARE_ROW_EXCLUSIVE = 'SHARE ROW EXCLUSIVE'.freeze
       SHARE_UPDATE_EXCLUSIVE = 'SHARE UPDATE EXCLUSIVE'.freeze
       SQL_WITH_RECURSIVE = "WITH RECURSIVE ".freeze
+      UPDATE_CLAUSE_METHODS = Dataset.clause_methods(:update, %w'table set from where')
       
       # Shared methods for prepared statements when used with PostgreSQL databases.
       module PreparedStatementMethods
@@ -625,6 +674,17 @@ module Sequel
         explain(:analyze=>true)
       end
       
+      # Handle converting the ruby xor operator (^) into the
+      # PostgreSQL xor operator (#).
+      def complex_expression_sql(op, args)
+        case op
+        when :^
+          "(#{literal(args.at(0))} # #{literal(args.at(1))})"
+        else
+          super
+        end
+      end
+
       # Disable the use of INSERT RETURNING, even if the server supports it
       def disable_insert_returning
         clone(:disable_insert_returning=>true)
@@ -635,14 +695,9 @@ module Sequel
         with_sql((opts[:analyze] ? EXPLAIN_ANALYZE : EXPLAIN) + select_sql).map(QUERY_PLAN).join("\r\n")
       end
       
-      # Return a cloned dataset with a :share lock type.
+      # Return a cloned dataset which will use FOR SHARE to lock returned rows.
       def for_share
-        clone(:lock => :share)
-      end
-      
-      # Return a cloned dataset with a :update lock type.
-      def for_update
-        clone(:lock => :update)
+        lock_style(:share)
       end
       
       # PostgreSQL specific full text search syntax, using tsearch2 (included
@@ -696,6 +751,16 @@ module Sequel
         [insert_sql(columns, LiteralString.new('VALUES ' + values.map {|r| literal(Array(r))}.join(COMMA_SEPARATOR)))]
       end
       
+      # DISTINCT ON is a PostgreSQL extension
+      def supports_distinct_on?
+        true
+      end
+      
+      # PostgreSQL supports modifying joined datasets
+      def supports_modifying_joins?
+        true
+      end
+
       # PostgreSQL supports timezones in literal timestamps
       def supports_timestamp_timezones?
         true
@@ -713,15 +778,41 @@ module Sequel
       
       private
       
+      # PostgreSQL allows deleting from joined datasets
+      def delete_clause_methods
+        DELETE_CLAUSE_METHODS
+      end 
+
+      # Only include the primary table in the main delete clause
+      def delete_from_sql(sql)
+        sql << " FROM #{source_list(@opts[:from][0..0])}"
+      end
+
+      # Use USING to specify additional tables in a delete query
+      def delete_using_sql(sql)
+        join_from_sql(:USING, sql)
+      end
+
       # Use the RETURNING clause to return the primary key of the inserted record, if it exists
       def insert_returning_pk_sql(*values)
         pk = db.primary_key(opts[:from].first) if opts[:from] && !opts[:from].empty?
         insert_returning_sql(pk ? Sequel::SQL::Identifier.new(pk) : NULL, *values)
       end
       
+      # For multiple table support, PostgreSQL requires at least
+      # two from tables, with joins allowed.
+      def join_from_sql(type, sql)
+        if(from = @opts[:from][1..-1]).empty?
+          raise(Error, 'Need multiple FROM tables if updating/deleting a dataset with JOINs') if @opts[:join]
+        else
+          sql << " #{type} #{source_list(from)}"
+          select_join_sql(sql)
+        end
+      end
+
       # Use a generic blob quoting method, hopefully overridden in one of the subadapter methods
       def literal_blob(v)
-        "'#{v.gsub(/[\000-\037\047\134\177-\377]/){|b| "\\#{("%o" % b[0..1].unpack("C")[0]).rjust(3, '0')}"}}'"
+        "'#{v.gsub(/[\000-\037\047\134\177-\377]/n){|b| "\\#{("%o" % b[0..1].unpack("C")[0]).rjust(3, '0')}"}}'"
       end
 
       # PostgreSQL uses FALSE for false values
@@ -743,21 +834,16 @@ module Sequel
       def select_clause_methods
         server_version >= 80400 ? SELECT_CLAUSE_METHODS_84 : SELECT_CLAUSE_METHODS
       end
+      
+      # Support FOR SHARE locking when using the :share lock style.
+      def select_lock_sql(sql)
+        @opts[:lock] == :share ? (sql << FOR_SHARE) : super
+      end
 
       # SQL fragment for named window specifications
       def select_window_sql(sql)
         sql << " WINDOW #{@opts[:window].map{|name, window| "#{literal(name)} AS #{literal(window)}"}.join(', ')}" if @opts[:window]
       end
-
-      # Support lock mode, allowing FOR SHARE and FOR UPDATE queries.
-      def select_lock_sql(sql)
-        case @opts[:lock]
-        when :update
-          sql << FOR_UPDATE
-        when :share
-          sql << FOR_SHARE
-        end
-      end
       
       # Use WITH RECURSIVE instead of WITH if any of the CTEs is recursive
       def select_with_sql_base
@@ -776,6 +862,21 @@ module Sequel
         cols.pop
         literal(SQL::StringExpression.new(:'||', *cols))
       end
+
+      # PostgreSQL splits the main table from the joined tables
+      def update_clause_methods
+        UPDATE_CLAUSE_METHODS
+      end
+
+      # Use FROM to specify additional tables in an update query
+      def update_from_sql(sql)
+        join_from_sql(:FROM, sql)
+      end
+
+      # Only include the primary table in the main update clause
+      def update_table_sql(sql)
+        sql << " #{source_list(@opts[:from][0..0])}"
+      end
     end
   end
 end
