@@ -42,6 +42,11 @@ module Sequel
       # plugins) in connection with option to check for typecast failures for
       # columns that aren't blobs or strings.
       attr_accessor :raise_on_typecast_failure
+      
+      # Whether to raise an error if an UPDATE or DELETE query related to
+      # a model instance does not modify exactly 1 row.  If set to false,
+      # Sequel will not check the number of rows modified (default: true).
+      attr_accessor :require_modification
   
       # Which columns are specifically restricted in a call to set/update/new/etc.
       # (default: not set).  Some columns are restricted regardless of
@@ -108,6 +113,11 @@ module Sequel
       def dataset
         @dataset || raise(Error, "No dataset associated with #{self}")
       end
+
+      # Alias of set_dataset
+      def dataset=(ds)
+        set_dataset(ds)
+      end
     
       # Returns the database associated with the Model class.
       # If this model doesn't have a database associated with it,
@@ -165,8 +175,8 @@ module Sequel
       
       # Like find but invokes create with given conditions when record does not
       # exist.
-      def find_or_create(cond)
-        find(cond) || create(cond)
+      def find_or_create(cond, &block)
+        find(cond) || create(cond, &block)
       end
     
       # If possible, set the dataset for the model subclass as soon as it
@@ -231,6 +241,14 @@ module Sequel
           {key => value}
         end
       end
+
+      # Return a hash where the keys are qualified column references.  Uses the given
+      # qualifier if provided, or the table_name otherwise.
+      def qualified_primary_key_hash(value, qualifier=table_name)
+        h = primary_key_hash(value)
+        h.to_a.each{|k,v| h[SQL::QualifiedIdentifier.new(qualifier, k)] = h.delete(k)}
+        h
+      end
   
       # Restrict the setting of the primary key(s) inside new/set/update.  Because
       # this is the default, this only make sense to use in a subclass where the
@@ -282,6 +300,7 @@ module Sequel
           raise(Error, "Model.set_dataset takes a Symbol or a Sequel::Dataset")
         end
         @dataset.row_proc = Proc.new{|r| load(r)}
+        @require_modification = Sequel::Model.require_modification.nil? ? @dataset.provides_accurate_rows_matched? : Sequel::Model.require_modification
         if inherited
           @simple_table = superclass.simple_table
           @columns = @dataset.columns rescue nil
@@ -290,16 +309,9 @@ module Sequel
           @dataset_methods.each{|meth, block| @dataset.meta_def(meth, &block)} if @dataset_methods
         end
         @dataset.model = self if @dataset.respond_to?(:model=)
-        begin
-          @db_schema = (inherited ? superclass.db_schema : get_db_schema)
-        rescue Sequel::DatabaseConnectionError
-          raise
-        rescue
-          nil
-        end
+        check_non_connection_error{@db_schema = (inherited ? superclass.db_schema : get_db_schema)}
         self
       end
-      alias dataset= set_dataset
     
       # Sets the primary key for this model. You can use either a regular 
       # or a composite primary key.
@@ -307,7 +319,7 @@ module Sequel
       # Example:
       #   class Tagging < Sequel::Model
       #     # composite key
-      #     set_primary_key :taggable_id, :tag_id
+      #     set_primary_key [:taggable_id, :tag_id]
       #   end
       #
       #   class Person < Sequel::Model
@@ -368,6 +380,17 @@ module Sequel
   
       private
       
+      # Yield to the passed block and swallow all errors other than DatabaseConnectionErrors.
+      def check_non_connection_error
+        begin
+          yield
+        rescue Sequel::DatabaseConnectionError
+          raise
+        rescue
+          nil
+        end
+      end
+
       # Create a column accessor for a column with a method name that is hard to use in ruby code.
       def def_bad_column_accessor(column)
         overridable_methods_module.module_eval do
@@ -399,15 +422,7 @@ module Sequel
         ds_opts = dataset.opts
         single_table = ds_opts[:from] && (ds_opts[:from].length == 1) \
           && !ds_opts.include?(:join) && !ds_opts.include?(:sql)
-        get_columns = proc do
-          begin
-            columns
-          rescue Sequel::DatabaseConnectionError
-            raise
-          rescue
-            []
-          end
-        end
+        get_columns = proc{check_non_connection_error{columns} || []}
         if single_table && (schema_array = (db.schema(table_name, :reload=>reload) rescue nil))
           schema_array.each{|k,v| schema_hash[k] = v}
           if ds_opts.include?(:select)
@@ -440,6 +455,21 @@ module Sequel
         end
         schema_hash
       end
+      
+      # For the given opts hash and default name or :class option, add a
+      # :class_name option unless already present which contains the name
+      # of the class to use as a string.  The purpose is to allow late
+      # binding to the class later using constantize.
+      def late_binding_class_option(opts, default)
+        case opts[:class]
+          when String, Symbol
+            # Delete :class to allow late binding
+            opts[:class_name] ||= opts.delete(:class).to_s
+          when Class
+            opts[:class_name] ||= opts[:class].name
+        end
+        opts[:class_name] ||= ((name || '').split("::")[0..-2] + [camelize(default)]).join('::')
+      end
   
       # Module that the class includes that holds methods the class adds for column accessors and
       # associations so that the methods can be overridden with super
@@ -481,7 +511,7 @@ module Sequel
     # * The following instance_methods all call the class method of the same
     #   name: columns, dataset, db, primary_key, db_schema.
     # * The following instance methods allow boolean flags to be set on a per-object
-    #   basis: raise_on_save_failure, raise_on_typecast_failure, strict_param_setting, 
+    #   basis: raise_on_save_failure, raise_on_typecast_failure, require_modification, strict_param_setting, 
     #   typecast_empty_string_to_nil, typecast_on_assignment, use_transactions.
     #   If they are not used, the object will default to whatever the model setting is.
     module InstanceMethods
@@ -506,8 +536,8 @@ module Sequel
       private_class_method :class_attr_overridable, :class_attr_reader
 
       class_attr_reader :columns, :db, :primary_key, :db_schema
-      class_attr_overridable :raise_on_save_failure, :raise_on_typecast_failure, :strict_param_setting, :typecast_empty_string_to_nil, :typecast_on_assignment, :use_transactions
-      
+      class_attr_overridable *BOOLEAN_SETTINGS
+
       # The hash of attribute values.  Keys are symbols with the names of the
       # underlying database columns.
       attr_reader :values
@@ -558,11 +588,10 @@ module Sequel
         end
       end
   
-      # Compares model instances by values.
+      # Alias of eql?
       def ==(obj)
-        (obj.class == model) && (obj.values == @values)
+        eql?(obj)
       end
-      alias eql? ==
   
       # If pk is not nil, true only if the objects have the same class and pk.
       # If pk is nil, false.
@@ -575,13 +604,6 @@ module Sequel
       # this alias makes it so you can use model instead of
       # self.class.
       alias_method :model, :class
-  
-      # The current cached associations.  A hash with the keys being the
-      # association name symbols and the values being the associated object
-      # or nil (many_to_one), or the array of associated objects (*_to_many).
-      def associations
-        @associations ||= {}
-      end
 
       # The autoincrementing primary key for this model object. Should be
       # overridden if you have a composite primary key with one part of it
@@ -599,7 +621,7 @@ module Sequel
       # Deletes and returns self.  Does not run destroy hooks.
       # Look into using destroy instead.
       def delete
-        this.delete
+        _delete
         self
       end
       
@@ -607,9 +629,10 @@ module Sequel
       # If before_destroy returns false, returns false without
       # deleting the object the the database. Otherwise, deletes
       # the item from the database and returns self.  Uses a transaction
-      # if use_transactions is true.
-      def destroy
-        use_transactions ? db.transaction{_destroy} : _destroy
+      # if use_transactions is true or if the :transaction option is given and
+      # true.
+      def destroy(opts = {})
+        checked_save_failure{checked_transaction(opts){_destroy(opts)}}
       end
 
       # Iterates through all of the current values using each.
@@ -620,6 +643,11 @@ module Sequel
         @values.each(&block)
       end
   
+      # Compares model instances by values.
+      def eql?(obj)
+        (obj.class == model) && (obj.values == @values)
+      end
+
       # Returns the validation errors associated with this object.
       def errors
         @errors ||= Errors.new
@@ -655,6 +683,11 @@ module Sequel
         @values.keys
       end
       
+      # Refresh this record using for_update unless this is a new record.  Returns self.
+      def lock!
+        new? ? self : _refresh(this.for_update)
+      end
+      
       # Remove elements of the model object that make marshalling fail. Returns self.
       def marshallable!
         @this = nil
@@ -721,9 +754,11 @@ module Sequel
       # * :validate - set to false not to validate the model before saving
       def save(*columns)
         opts = columns.last.is_a?(Hash) ? columns.pop : {}
-        return save_failure(:invalid) if opts[:validate] != false and !valid?
-        use_transaction = opts.include?(:transaction) ? opts[:transaction] : use_transactions
-        use_transaction ? db.transaction(opts){_save(columns, opts)} : _save(columns, opts)
+        if opts[:validate] != false and !valid?
+          raise(ValidationFailed.new(errors)) if raise_on_save_failure
+          return
+        end
+        checked_save_failure{checked_transaction(opts){_save(columns, opts)}}
       end
 
       # Saves only changed columns if the object has been modified.
@@ -753,6 +788,13 @@ module Sequel
         set_restricted(hash, false, except.flatten)
       end
   
+      # For each of the fields in the given array +fields+, call the setter
+      # method with the value of that +hash+ entry for the field. Returns self.
+      def set_fields(hash, fields)
+        fields.each{|f| send("#{f}=", hash[f])}
+        self
+      end
+  
       # Set the values using the entries in the hash, only if the key
       # is included in only.
       def set_only(hash, *only)
@@ -781,6 +823,13 @@ module Sequel
         update_restricted(hash, false, except.flatten)
       end
   
+      # Update the instances values by calling +set_fields+ with the +hash+
+      # and +fields+, then save any changes to the record.  Returns self.
+      def update_fields(hash, fields)
+        set_fields(hash, fields)
+        save_changes
+      end
+
       # Update the values using the entries in the hash, only if the key
       # is included in only.
       def update_only(hash, *only)
@@ -789,7 +838,8 @@ module Sequel
       
       # Validates the object.  If the object is invalid, errors should be added
       # to the errors attribute.  By default, does nothing, as all models
-      # are valid by default.
+      # are valid by default.  See the {"Model Validations" guide}[link:files/doc/validations_rdoc.html].
+      # for details about validation.
       def validate
       end
 
@@ -797,7 +847,7 @@ module Sequel
       def valid?
         errors.clear
         if before_validation == false
-          save_failure(:validation)
+          save_failure(:validation) if raise_on_save_failure
           return false
         end
         validate
@@ -806,18 +856,40 @@ module Sequel
       end
 
       private
+      
+      # Actually do the deletion of the object's dataset.
+      def _delete
+        n = _delete_dataset.delete
+        raise(NoExistingObject, "Attempt to delete object did not result in a single row modification (Rows Deleted: #{n}, SQL: #{_delete_dataset.delete_sql})") if require_modification && n != 1
+        n
+      end
+      
+      # The dataset to use when deleting the object.   The same as the object's
+      # dataset by default.
+      def _delete_dataset
+        this
+      end
   
       # Internal destroy method, separted from destroy to
       # allow running inside a transaction
-      def _destroy
+      def _destroy(opts)
         return save_failure(:destroy) if before_destroy == false
-        delete
+        _destroy_delete
         after_destroy
         self
       end
       
+      # Internal delete method to call when destroying an object,
+      # separated from delete to allow you to override destroy's version
+      # without affecting delete.
+      def _destroy_delete
+        delete
+      end
+
+      # Insert the record into the database, returning the primary key if
+      # the record should be refreshed from the database.
       def _insert
-        ds = model.dataset
+        ds = _insert_dataset
         if ds.respond_to?(:insert_select) and h = ds.insert_select(@values)
           @values = h
           nil
@@ -832,12 +904,17 @@ module Sequel
         end
       end
       
+      # The dataset to use when inserting a new object.   The same as the model's
+      # dataset by default.
+      def _insert_dataset
+        model.dataset
+      end
+  
       # Refresh using a particular dataset, used inside save to make sure the same server
       # is used for reading newly inserted values from the database
       def _refresh(dataset)
         set_values(dataset.first || raise(Error, "Record not found"))
         changed_columns.clear
-        associations.clear
         self
       end
       
@@ -848,30 +925,28 @@ module Sequel
         if new?
           return save_failure(:create) if before_create == false
           pk = _insert
-          @this = nil if pk
+          @this = nil
           @new = false
           @was_new = true
           after_create
           after_save
           @was_new = nil
-          if pk
-            ds = this
-            ds = ds.server(:default) unless ds.opts[:server]
-            _refresh(ds)
-          else
-            changed_columns.clear
-          end
+          pk ? _save_refresh : changed_columns.clear
         else
           return save_failure(:update) if before_update == false
           if columns.empty?
-            @columns_updated = opts[:changed] ? @values.reject{|k,v| !changed_columns.include?(k)} : @values.dup
+            @columns_updated = if opts[:changed]
+              @values.reject{|k,v| !changed_columns.include?(k)}
+            else
+              _save_update_all_columns_hash
+            end
             changed_columns.clear
           else # update only the specified columns
             @columns_updated = @values.reject{|k, v| !columns.include?(k)}
             changed_columns.reject!{|c| columns.include?(c)}
           end
-          Array(primary_key).each{|x| @columns_updated.delete(x)}
           _update(@columns_updated) unless @columns_updated.empty?
+          @this = nil
           after_update
           after_save
           @columns_updated = nil
@@ -879,11 +954,58 @@ module Sequel
         @modified = false
         self
       end
-      
-      def _update(columns)
-        this.update(columns)
+
+      # Refresh the object after saving it, used to get
+      # default values of all columns.  Separated from _save so it
+      # can be overridden to avoid the refresh.
+      def _save_refresh
+        _refresh(this.opts[:server] ? this : this.server(:default))
+      end
+
+      # Return a hash of values used when saving all columns of an
+      # existing object (i.e. not passing specific columns to save
+      # or using update/save_changes).  Defaults to all of the
+      # object's values except unmodified primary key columns, as some
+      # databases don't like you setting primary key values even
+      # to their existing values.
+      def _save_update_all_columns_hash
+        v = @values.dup
+        Array(primary_key).each{|x| v.delete(x) unless changed_columns.include?(x)}
+        v
       end
       
+      # Update this instance's dataset with the supplied column hash.
+      def _update(columns)
+        n = _update_dataset.update(columns)
+        raise(NoExistingObject, "Attempt to update object did not result in a single row modification (SQL: #{_update_dataset.update_sql(columns)})") if require_modification && n != 1
+        n
+      end
+      
+      # The dataset to use when updating an object.  The same as the object's
+      # dataset by default.
+      def _update_dataset
+        this
+      end
+
+      # If raise_on_save_failure is false, check for BeforeHookFailed
+      # beind raised by yielding and swallow it.
+      def checked_save_failure
+        if raise_on_save_failure
+          yield
+        else
+          begin
+            yield
+          rescue BeforeHookFailed 
+            nil
+          end
+        end
+      end
+      
+      # If transactions should be used, wrap the yield in a transaction block.
+      def checked_transaction(opts={})
+        use_transaction?(opts) ? db.transaction(opts){yield} : yield
+      end
+
       # Default inspection output for the values hash, overwrite to change what #inspect displays.
       def inspect_values
         @values.inspect
@@ -891,13 +1013,7 @@ module Sequel
   
       # Raise an error if raise_on_save_failure is true, return nil otherwise.
       def save_failure(type)
-        if raise_on_save_failure
-          if type == :invalid
-            raise ValidationFailed, errors.full_messages.join(', ')
-          else
-            raise BeforeHookFailed, "one of the before_#{type} hooks returned false"
-          end
-        end
+        raise BeforeHookFailed, "one of the before_#{type} hooks returned false"
       end
   
       # Set the columns, filtered by the only and except arrays.
@@ -968,6 +1084,13 @@ module Sequel
         set_restricted(hash, only, except)
         save_changes
       end
+      
+      # Whether to use a transaction for this action.  If the :transaction
+      # option is present in the hash, use that, otherwise, fallback to the
+      # object's default (if set), or class's default (if not).
+      def use_transaction?(opts = {})
+        opts.fetch(:transaction, use_transactions)
+      end
     end
 
     # Dataset methods are methods that the model class extends its dataset with in
