@@ -3,11 +3,11 @@ module Sequel
     # Errors represents validation errors, a simple hash subclass
     # with a few convenience methods.
     class Errors < ::Hash
-      ATTRIBUTE_JOINER = ' and '
+      ATTRIBUTE_JOINER = ' and '.freeze
 
       # Assign an array of messages for each attribute on access
-      def initialize
-        super{|h,k| h[k] = []}
+      def [](k)
+        has_key?(k) ? super : (self[k] = [])
       end
 
       # Adds an error for the given attribute.
@@ -19,6 +19,11 @@ module Sequel
       def count
         values.inject(0){|m, v| m + v.length}
       end
+       
+      # Return true if there are no error messages, false otherwise.
+      def empty?
+        count == 0
+      end
       
       # Returns an array of fully-formatted error messages.
       def full_messages
@@ -32,7 +37,9 @@ module Sequel
       # Returns the array of errors for the given attribute, or nil
       # if there are no errors for the attribute.
       def on(att)
-        self[att] if include?(att)
+        if v = fetch(att, nil) and !v.empty?
+          v
+        end
       end
     end
   end
