@@ -1,15 +1,35 @@
 require 'rubygems'
+require 'logger'
 unless Object.const_defined?('Sequel')
-  $:.unshift(File.join(File.dirname(__FILE__), "../../lib/"))
+  $:.unshift(File.join(File.dirname(File.expand_path(__FILE__)), "../../lib/"))
   require 'sequel'
   Sequel.quote_identifiers = false
 end
 begin
-  require File.join(File.dirname(File.dirname(__FILE__)), 'spec_config.rb')
+  require File.join(File.dirname(File.dirname(File.expand_path(__FILE__))), 'spec_config.rb')
 rescue LoadError
 end
 
+class Sequel::Database
+  def log_duration(duration, message)
+    log_info(message)
+  end
+end
+
 class Spec::Example::ExampleGroup
+  def log 
+    begin
+      INTEGRATION_DB.loggers << Logger.new(STDOUT)
+      yield
+    ensure
+     INTEGRATION_DB.loggers.pop
+    end 
+  end 
+
+  def self.log_specify(message, &block)
+    specify(message){log{instance_eval(&block)}}
+  end
+
   def self.cspecify(message, *checked, &block)
     pending = false
     checked.each do |c|
@@ -28,4 +48,4 @@ class Spec::Example::ExampleGroup
       specify(message, &block)
     end
   end
-end
\ No newline at end of file
+end
