@@ -1,4 +1,4 @@
-require File.join(File.dirname(__FILE__), 'spec_helper')
+require File.join(File.dirname(File.expand_path(__FILE__)), 'spec_helper')
 
 context "A new Database" do
   before do
@@ -11,7 +11,8 @@ context "A new Database" do
   end
   
   specify "should receive options" do
-    @db.opts.should == {1 => 2, :logger => 3}  
+    @db.opts[1].should == 2
+    @db.opts[:logger].should == 3  
   end
   
   specify "should set the logger from opts[:logger] and opts[:loggers]" do
@@ -37,9 +38,17 @@ context "A new Database" do
   end
 
   specify "should respect the :single_threaded option" do
-    db = Sequel::Database.new(:single_threaded=>true)
-    db.pool.should be_a_kind_of(Sequel::SingleThreadedPool)
-    db = Sequel::Database.new(:single_threaded=>false)
+    db = Sequel::Database.new(:single_threaded=>true){123}
+    db.pool.should be_a_kind_of(Sequel::SingleConnectionPool)
+    db = Sequel::Database.new(:single_threaded=>'t'){123}
+    db.pool.should be_a_kind_of(Sequel::SingleConnectionPool)
+    db = Sequel::Database.new(:single_threaded=>'1'){123}
+    db.pool.should be_a_kind_of(Sequel::SingleConnectionPool)
+    db = Sequel::Database.new(:single_threaded=>false){123}
+    db.pool.should be_a_kind_of(Sequel::ConnectionPool)
+    db = Sequel::Database.new(:single_threaded=>'f'){123}
+    db.pool.should be_a_kind_of(Sequel::ConnectionPool)
+    db = Sequel::Database.new(:single_threaded=>'0'){123}
     db.pool.should be_a_kind_of(Sequel::ConnectionPool)
   end
 
@@ -183,14 +192,104 @@ context "Database#disconnect" do
   specify "should call pool.disconnect" do
     d = Sequel::Database.new
     p = d.pool
-    p.should_receive(:disconnect).once.and_return(2)
+    p.should_receive(:disconnect).once.with({}).and_return(2)
     d.disconnect.should == 2
   end
 end
 
+context "Sequel.extension" do
+  specify "should attempt to load the given extension" do
+    proc{Sequel.extension :blah}.should raise_error(LoadError)
+  end
+end
+
 context "Database#connect" do
-  specify "should raise NotImplementedError" do
-    proc {Sequel::Database.new.connect}.should raise_error(NotImplementedError)
+  specify "should raise Sequel::NotImplemented" do
+    proc {Sequel::Database.new.connect(:default)}.should raise_error(Sequel::NotImplemented)
+  end
+end
+
+context "Database#log_info" do
+  before do
+    @o = Object.new
+    def @o.logs; @logs || []; end
+    def @o.to_ary; [self]; end
+    def @o.method_missing(*args); (@logs ||= []) << args; end
+    @db = Sequel::Database.new(:logger=>@o)
+  end
+
+  specify "should log message at info level to all loggers" do
+    @db.log_info('blah')
+    @o.logs.should == [[:info, 'blah']]
+  end
+
+  specify "should log message with args at info level to all loggers" do
+    @db.log_info('blah', [1, 2])
+    @o.logs.should == [[:info, 'blah; [1, 2]']]
+  end
+end
+
+context "Database#log_yield" do
+  before do
+    @o = Object.new
+    def @o.logs; @logs || []; end
+    def @o.warn(*args); (@logs ||= []) << [:warn] + args; end
+    def @o.method_missing(*args); (@logs ||= []) << args; end
+    def @o.to_ary; [self]; end
+    @db = Sequel::Database.new(:logger=>@o)
+  end
+
+  specify "should yield to the passed block" do
+    a = nil
+    @db.log_yield('blah'){a = 1}
+    a.should == 1
+  end
+
+  specify "should raise an exception if a block is not passed" do
+    proc{@db.log_yield('blah')}.should raise_error
+  end
+
+  specify "should log message with duration at info level to all loggers" do
+    @db.log_yield('blah'){}
+    @o.logs.length.should == 1
+    @o.logs.first.length.should == 2
+    @o.logs.first.first.should == :info
+    @o.logs.first.last.should =~ /\A\(\d\.\d{6}s\) blah\z/ 
+  end
+
+  specify "should log message with duration at warn level if duration greater than log_warn_duration" do
+    @db.log_warn_duration = 0
+    @db.log_yield('blah'){}
+    @o.logs.length.should == 1
+    @o.logs.first.length.should == 2
+    @o.logs.first.first.should == :warn
+    @o.logs.first.last.should =~ /\A\(\d\.\d{6}s\) blah\z/ 
+  end
+
+  specify "should log message with duration at info level if duration less than log_warn_duration" do
+    @db.log_warn_duration = 1000
+    @db.log_yield('blah'){}
+    @o.logs.length.should == 1
+    @o.logs.first.length.should == 2
+    @o.logs.first.first.should == :info
+    @o.logs.first.last.should =~ /\A\(\d\.\d{6}s\) blah\z/ 
+  end
+
+  specify "should log message at error level if block raises an error" do
+    @db.log_warn_duration = 0
+    proc{@db.log_yield('blah'){raise Sequel::Error, 'adsf'}}.should raise_error
+    @o.logs.length.should == 1
+    @o.logs.first.length.should == 2
+    @o.logs.first.first.should == :error
+    @o.logs.first.last.should =~ /\ASequel::Error: adsf: blah\z/ 
+  end
+
+  specify "should include args with message if args passed" do
+    @db.log_yield('blah', [1, 2]){}
+    @o.logs.length.should == 1
+    @o.logs.first.length.should == 2
+    @o.logs.first.first.should == :info
+    @o.logs.first.last.should =~ /\A\(\d\.\d{6}s\) blah; \[1, 2\]\z/ 
   end
 end
 
@@ -266,34 +365,54 @@ context "Database#dataset" do
 end
 
 context "Database#execute" do
-  specify "should raise NotImplementedError" do
-    proc {Sequel::Database.new.execute('blah blah')}.should raise_error(NotImplementedError)
-    proc {Sequel::Database.new << 'blah blah'}.should raise_error(NotImplementedError)
+  specify "should raise Sequel::NotImplemented" do
+    proc {Sequel::Database.new.execute('blah blah')}.should raise_error(Sequel::NotImplemented)
+    proc {Sequel::Database.new << 'blah blah'}.should raise_error(Sequel::NotImplemented)
   end
 end
 
-context "Database#<<" do
+context "Database#tables" do
+  specify "should raise Sequel::NotImplemented" do
+    proc {Sequel::Database.new.tables}.should raise_error(Sequel::NotImplemented)
+  end
+end
+
+context "Database#indexes" do
+  specify "should raise Sequel::NotImplemented" do
+    proc {Sequel::Database.new.indexes(:table)}.should raise_error(Sequel::NotImplemented)
+  end
+end
+
+context "Database#<< and run" do
   before do
+    sqls = @sqls = []
     @c = Class.new(Sequel::Database) do
-      define_method(:execute) {|sql, opts| sql}
+      define_method(:execute_ddl){|sql, *opts| sqls.clear; sqls << sql; sqls.concat(opts)}
     end
     @db = @c.new({})
   end
   
-  specify "should pass the supplied sql to #execute" do
-    (@db << "DELETE FROM items").should == "DELETE FROM items"
+  specify "should pass the supplied sql to #execute_ddl" do
+    (@db << "DELETE FROM items")
+    @sqls.should == ["DELETE FROM items", {}]
+    @db.run("DELETE FROM items2")
+    @sqls.should == ["DELETE FROM items2", {}]
   end
   
-  specify "should not remove comments and whitespace from strings" do
-    s = "INSERT INTO items VALUES ('---abc')"
-    (@db << s).should == s
+  specify "should return nil" do
+    (@db << "DELETE FROM items").should == nil
+    @db.run("DELETE FROM items").should == nil
+  end
+  
+  specify "should accept options passed to execute_ddl" do
+    @db.run("DELETE FROM items", :server=>:s1)
+    @sqls.should == ["DELETE FROM items", {:server=>:s1}]
   end
 end
 
 context "Database#synchronize" do
   before do
-    @db = Sequel::Database.new(:max_connections => 1)
-    @db.pool.connection_proc = proc {12345}
+    @db = Sequel::Database.new(:max_connections => 1){12345}
   end
   
   specify "should wrap the supplied block in pool.hold" do
@@ -316,12 +435,10 @@ end
 
 context "Database#test_connection" do
   before do
-    @db = Sequel::Database.new
-    @test = nil
-    @db.pool.connection_proc = proc {@test = rand(100)}
+    @db = Sequel::Database.new{@test = rand(100)}
   end
   
-  specify "should call pool#hold" do
+  specify "should pool#hold" do
     @db.test_connection
     @test.should_not be_nil
   end
@@ -329,6 +446,10 @@ context "Database#test_connection" do
   specify "should return true if successful" do
     @db.test_connection.should be_true
   end
+
+  specify "should raise an error if the attempting to connect raises an error" do
+    proc{Sequel::Database.new{raise Sequel::Error, 'blah'}.test_connection}.should raise_error(Sequel::Error)
+  end
 end
 
 class DummyDataset < Sequel::Dataset
@@ -551,19 +672,10 @@ context "Database#rename_table" do
 end
 
 context "Database#table_exists?" do
-  before do
-    @db = DummyDatabase.new
-    @db.instance_variable_set(:@schemas, {:a=>[]})
-    @db2 = DummyDatabase.new
-  end
-  
-  specify "should use schema information if available" do
-    @db.table_exists?(:a).should be_true
-  end
-  
-  specify "should otherwise try to select the first record from the table's dataset" do
-    @db2.table_exists?(:a).should be_false
-    @db2.table_exists?(:b).should be_true
+  specify "should try to select the first record from the table's dataset" do
+    db2 = DummyDatabase.new
+    db2.table_exists?(:a).should be_false
+    db2.table_exists?(:b).should be_true
   end
 end
 
@@ -579,15 +691,37 @@ end
 
 context "Database#transaction" do
   before do
-    @db = Dummy3Database.new
-    @db.pool.connection_proc = proc {Dummy3Database::DummyConnection.new(@db)}
+    @db = Dummy3Database.new{Dummy3Database::DummyConnection.new(@db)}
   end
   
   specify "should wrap the supplied block with BEGIN + COMMIT statements" do
-    @db.transaction {@db.execute 'DROP TABLE test;'}
+    @db.transaction{@db.execute 'DROP TABLE test;'}
     @db.sql.should == ['BEGIN', 'DROP TABLE test;', 'COMMIT']
   end
   
+  specify "should support transaction isolation levels" do
+    @db.meta_def(:supports_transaction_isolation_levels?){true}
+    [:uncommitted, :committed, :repeatable, :serializable].each do |l|
+      @db.transaction(:isolation=>l){@db.run "DROP TABLE #{l}"}
+    end
+    @db.sql.should == ['BEGIN', 'SET TRANSACTION ISOLATION LEVEL READ UNCOMMITTED', 'DROP TABLE uncommitted', 'COMMIT',
+                       'BEGIN', 'SET TRANSACTION ISOLATION LEVEL READ COMMITTED', 'DROP TABLE committed', 'COMMIT',
+                       'BEGIN', 'SET TRANSACTION ISOLATION LEVEL REPEATABLE READ', 'DROP TABLE repeatable', 'COMMIT',
+                       'BEGIN', 'SET TRANSACTION ISOLATION LEVEL SERIALIZABLE', 'DROP TABLE serializable', 'COMMIT']
+  end
+
+  specify "should allow specifying a default transaction isolation level" do
+    @db.meta_def(:supports_transaction_isolation_levels?){true}
+    [:uncommitted, :committed, :repeatable, :serializable].each do |l|
+      @db.transaction_isolation_level = l
+      @db.transaction{@db.run "DROP TABLE #{l}"}
+    end
+    @db.sql.should == ['BEGIN', 'SET TRANSACTION ISOLATION LEVEL READ UNCOMMITTED', 'DROP TABLE uncommitted', 'COMMIT',
+                       'BEGIN', 'SET TRANSACTION ISOLATION LEVEL READ COMMITTED', 'DROP TABLE committed', 'COMMIT',
+                       'BEGIN', 'SET TRANSACTION ISOLATION LEVEL REPEATABLE READ', 'DROP TABLE repeatable', 'COMMIT',
+                       'BEGIN', 'SET TRANSACTION ISOLATION LEVEL SERIALIZABLE', 'DROP TABLE serializable', 'COMMIT']
+  end
+  
   specify "should handle returning inside of the block by committing" do
     def @db.ret_commit
       transaction do
@@ -617,6 +751,14 @@ context "Database#transaction" do
     @db.sql.should == ['BEGIN', 'DROP TABLE a', 'ROLLBACK']
   end
   
+  specify "should raise database errors when commiting a transaction as Sequel::DatabaseError" do
+    @db.meta_def(:commit_transaction){raise ArgumentError}
+    lambda{@db.transaction{}}.should raise_error(ArgumentError)
+
+    @db.meta_def(:database_error_classes){[ArgumentError]}
+    lambda{@db.transaction{}}.should raise_error(Sequel::DatabaseError)
+  end
+  
   specify "should be re-entrant" do
     stop = false
     cc = nil
@@ -635,6 +777,105 @@ context "Database#transaction" do
   end
 end
 
+context "Database#transaction with savepoints" do
+  before do
+    @db = Dummy3Database.new{Dummy3Database::DummyConnection.new(@db)}
+    @db.meta_def(:supports_savepoints?){true}
+  end
+  
+  specify "should wrap the supplied block with BEGIN + COMMIT statements" do
+    @db.transaction {@db.execute 'DROP TABLE test;'}
+    @db.sql.should == ['BEGIN', 'DROP TABLE test;', 'COMMIT']
+  end
+  
+  specify "should use savepoints if given the :savepoint option" do
+    @db.transaction{@db.transaction(:savepoint=>true){@db.execute 'DROP TABLE test;'}}
+    @db.sql.should == ['BEGIN', 'SAVEPOINT autopoint_1', 'DROP TABLE test;', 'RELEASE SAVEPOINT autopoint_1', 'COMMIT']
+  end
+  
+  specify "should not use a savepoints if no transaction is in progress" do
+    @db.transaction(:savepoint=>true){@db.execute 'DROP TABLE test;'}
+    @db.sql.should == ['BEGIN', 'DROP TABLE test;', 'COMMIT']
+  end
+  
+  specify "should reuse the current transaction if no :savepoint option is given" do
+    @db.transaction{@db.transaction{@db.execute 'DROP TABLE test;'}}
+    @db.sql.should == ['BEGIN', 'DROP TABLE test;', 'COMMIT']
+  end
+  
+  specify "should handle returning inside of the block by committing" do
+    def @db.ret_commit
+      transaction do
+        execute 'DROP TABLE test;'
+        return
+        execute 'DROP TABLE test2;';
+      end
+    end
+    @db.ret_commit
+    @db.sql.should == ['BEGIN', 'DROP TABLE test;', 'COMMIT']
+  end
+  
+  specify "should handle returning inside of a savepoint by committing" do
+    def @db.ret_commit
+      transaction do
+        transaction(:savepoint=>true) do
+          execute 'DROP TABLE test;'
+          return
+          execute 'DROP TABLE test2;';
+        end
+      end
+    end
+    @db.ret_commit
+    @db.sql.should == ['BEGIN', 'SAVEPOINT autopoint_1', 'DROP TABLE test;', 'RELEASE SAVEPOINT autopoint_1', 'COMMIT']
+  end
+  
+  specify "should issue ROLLBACK if an exception is raised, and re-raise" do
+    @db.transaction {@db.execute 'DROP TABLE test'; raise RuntimeError} rescue nil
+    @db.sql.should == ['BEGIN', 'DROP TABLE test', 'ROLLBACK']
+    
+    proc {@db.transaction {raise RuntimeError}}.should raise_error(RuntimeError)
+  end
+  
+  specify "should issue ROLLBACK SAVEPOINT if an exception is raised inside a savepoint, and re-raise" do
+    @db.transaction{@db.transaction(:savepoint=>true){@db.execute 'DROP TABLE test'; raise RuntimeError}} rescue nil
+    @db.sql.should == ['BEGIN', 'SAVEPOINT autopoint_1', 'DROP TABLE test', 'ROLLBACK TO SAVEPOINT autopoint_1', 'ROLLBACK']
+    
+    proc {@db.transaction {raise RuntimeError}}.should raise_error(RuntimeError)
+  end
+  
+  specify "should issue ROLLBACK if Sequel::Rollback is called in the transaction" do
+    @db.transaction do
+      @db.drop_table(:a)
+      raise Sequel::Rollback
+      @db.drop_table(:b)
+    end
+    
+    @db.sql.should == ['BEGIN', 'DROP TABLE a', 'ROLLBACK']
+  end
+  
+  specify "should issue ROLLBACK SAVEPOINT if Sequel::Rollback is called in a savepoint" do
+    @db.transaction do
+      @db.transaction(:savepoint=>true) do
+        @db.drop_table(:a)
+        raise Sequel::Rollback
+      end
+      @db.drop_table(:b)
+    end
+    
+    @db.sql.should == ['BEGIN', 'SAVEPOINT autopoint_1', 'DROP TABLE a', 'ROLLBACK TO SAVEPOINT autopoint_1', 'DROP TABLE b', 'COMMIT']
+  end
+  
+  specify "should raise database errors when commiting a transaction as Sequel::DatabaseError" do
+    @db.meta_def(:commit_transaction){raise ArgumentError}
+    lambda{@db.transaction{}}.should raise_error(ArgumentError)
+    lambda{@db.transaction{@db.transaction(:savepoint=>true){}}}.should raise_error(ArgumentError)
+
+    @db.meta_def(:database_error_classes){[ArgumentError]}
+    lambda{@db.transaction{}}.should raise_error(Sequel::DatabaseError)
+    lambda{@db.transaction{@db.transaction(:savepoint=>true){}}}.should raise_error(Sequel::DatabaseError)
+  end
+end
+
 context "A Database adapter with a scheme" do
   before do
     class CCC < Sequel::Database
@@ -676,6 +917,7 @@ context "A Database adapter with a scheme" do
     x = nil
     y = nil
     z = nil
+    returnValue = 'anything'
 
     p = proc do |c|
       c.should be_a_kind_of(CCC)
@@ -684,15 +926,16 @@ context "A Database adapter with a scheme" do
       z = y
       y = x
       x = c
+      returnValue
     end
-    Sequel::Database.connect('ccc://localhost/db', &p).should == nil
+    Sequel::Database.connect('ccc://localhost/db', &p).should == returnValue
     CCC::DISCONNECTS.should == [x]
 
-    Sequel.connect('ccc://localhost/db', &p).should == nil
+    Sequel.connect('ccc://localhost/db', &p).should == returnValue
     CCC::DISCONNECTS.should == [y, x]
 
     Sequel.send(:def_adapter_method, :ccc)
-    Sequel.ccc('db', :host=>'localhost', &p).should == nil
+    Sequel.ccc('db', :host=>'localhost', &p).should == returnValue
     CCC::DISCONNECTS.should == [z, y, x]
   end
 
@@ -703,36 +946,79 @@ context "A Database adapter with a scheme" do
     proc {Sequel.ccc('abc', 'def')}.should raise_error(Sequel::Error)
     
     c = Sequel.ccc('mydb')
+    p = proc{c.opts.delete_if{|k,v| k == :disconnection_proc || k == :single_threaded}}
     c.should be_a_kind_of(CCC)
-    c.opts.should == {:adapter=>:ccc, :database => 'mydb'}
+    p.call.should == {:adapter=>:ccc, :database => 'mydb'}
     
     c = Sequel.ccc('mydb', :host => 'localhost')
     c.should be_a_kind_of(CCC)
-    c.opts.should == {:adapter=>:ccc, :database => 'mydb', :host => 'localhost'}
+    p.call.should == {:adapter=>:ccc, :database => 'mydb', :host => 'localhost'}
     
     c = Sequel.ccc
     c.should be_a_kind_of(CCC)
-    c.opts.should == {:adapter=>:ccc}
+    p.call.should == {:adapter=>:ccc}
     
     c = Sequel.ccc(:database => 'mydb', :host => 'localhost')
     c.should be_a_kind_of(CCC)
-    c.opts.should == {:adapter=>:ccc, :database => 'mydb', :host => 'localhost'}
+    p.call.should == {:adapter=>:ccc, :database => 'mydb', :host => 'localhost'}
   end
   
   specify "should be accessible through Sequel.connect with options" do
     c = Sequel.connect(:adapter => :ccc, :database => 'mydb')
     c.should be_a_kind_of(CCC)
-    c.opts.should == {:adapter => :ccc, :database => 'mydb'}
+    c.opts[:adapter].should == :ccc
   end
 
   specify "should be accessible through Sequel.connect with URL parameters" do
-    c = Sequel.connect 'ccc://localhost/db?host=/tmp&user=test'
+    c = Sequel.connect 'ccc:///db?host=/tmp&user=test'
     c.should be_a_kind_of(CCC)
     c.opts[:host].should == '/tmp'
     c.opts[:database].should == 'db'
     c.opts[:user].should == 'test'
   end
+  
+  specify "should have URL parameters take precedence over fixed URL parts" do
+    c = Sequel.connect 'ccc://localhost/db?host=a&database=b'
+    c.should be_a_kind_of(CCC)
+    c.opts[:host].should == 'a'
+    c.opts[:database].should == 'b'
+  end
+  
+  specify "should have hash options take predence over URL parameters or parts" do
+    c = Sequel.connect 'ccc://localhost/db?host=/tmp', :host=>'a', :database=>'b', :user=>'c'
+    c.should be_a_kind_of(CCC)
+    c.opts[:host].should == 'a'
+    c.opts[:database].should == 'b'
+    c.opts[:user].should == 'c'
+  end
 
+  specify "should unescape values of URL parameters and parts" do
+    c = Sequel.connect 'ccc:///d%5bb%5d?host=domain%5cinstance'
+    c.should be_a_kind_of(CCC)
+    c.opts[:database].should == 'd[b]'
+    c.opts[:host].should == 'domain\\instance'
+  end
+
+  specify "should test the connection if test parameter is truthy" do
+    proc{Sequel.connect 'ccc:///d%5bb%5d?test=t'}.should raise_error(Sequel::DatabaseConnectionError)
+    proc{Sequel.connect 'ccc:///d%5bb%5d?test=1'}.should raise_error(Sequel::DatabaseConnectionError)
+    proc{Sequel.connect 'ccc:///d%5bb%5d', :test=>true}.should raise_error(Sequel::DatabaseConnectionError)
+    proc{Sequel.connect 'ccc:///d%5bb%5d', :test=>'t'}.should raise_error(Sequel::DatabaseConnectionError)
+  end
+
+  specify "should not test the connection if test parameter is not truthy" do
+    proc{Sequel.connect 'ccc:///d%5bb%5d?test=f'}.should_not raise_error
+    proc{Sequel.connect 'ccc:///d%5bb%5d?test=0'}.should_not raise_error
+    proc{Sequel.connect 'ccc:///d%5bb%5d', :test=>false}.should_not raise_error
+    proc{Sequel.connect 'ccc:///d%5bb%5d', :test=>'f'}.should_not raise_error
+  end
+end
+
+context "Sequel::Database.connect" do
+  specify "should raise an Error if not given a String or Hash" do
+    proc{Sequel::Database.connect(nil)}.should raise_error(Sequel::Error)
+    proc{Sequel::Database.connect(Object.new)}.should raise_error(Sequel::Error)
+  end
 end
 
 context "An unknown database scheme" do
@@ -765,26 +1051,26 @@ context "A single threaded database" do
     Sequel::Database.single_threaded = false
   end
   
-  specify "should use a SingleThreadedPool instead of a ConnectionPool" do
-    db = Sequel::Database.new(:single_threaded => true)
-    db.pool.should be_a_kind_of(Sequel::SingleThreadedPool)
+  specify "should use a SingleConnectionPool instead of a ConnectionPool" do
+    db = Sequel::Database.new(:single_threaded => true){123}
+    db.pool.should be_a_kind_of(Sequel::SingleConnectionPool)
   end
   
   specify "should be constructable using :single_threaded => true option" do
-    db = Sequel::Database.new(:single_threaded => true)
-    db.pool.should be_a_kind_of(Sequel::SingleThreadedPool)
+    db = Sequel::Database.new(:single_threaded => true){123}
+    db.pool.should be_a_kind_of(Sequel::SingleConnectionPool)
   end
   
   specify "should be constructable using Database.single_threaded = true" do
     Sequel::Database.single_threaded = true
-    db = Sequel::Database.new
-    db.pool.should be_a_kind_of(Sequel::SingleThreadedPool)
+    db = Sequel::Database.new{123}
+    db.pool.should be_a_kind_of(Sequel::SingleConnectionPool)
   end
 
   specify "should be constructable using Sequel.single_threaded = true" do
     Sequel.single_threaded = true
-    db = Sequel::Database.new
-    db.pool.should be_a_kind_of(Sequel::SingleThreadedPool)
+    db = Sequel::Database.new{123}
+    db.pool.should be_a_kind_of(Sequel::SingleConnectionPool)
   end
 end
 
@@ -801,12 +1087,24 @@ context "A single threaded database" do
     @db.pool.hold {|c| c.should == 1234568}
   end
   
-  specify "should convert an Exception into a RuntimeError" do
-    db = Sequel::Database.new(:single_threaded => true) do
-      raise Exception
-    end
-    
-    proc {db.pool.hold {|c|}}.should raise_error(RuntimeError)
+  specify "should disconnect correctly" do
+    def @db.disconnect_connection(c); @dc = c end
+    def @db.dc; @dc end
+    x = nil
+    @db.pool.hold{|c| x = c}
+    @db.pool.hold{|c| c.should == x}
+    proc{@db.disconnect}.should_not raise_error
+    @db.dc.should == x
+  end
+  
+  specify "should convert an Exception on connection into a DatabaseConnectionError" do
+    db = Sequel::Database.new(:single_threaded => true, :servers=>{}){raise Exception}
+    proc {db.pool.hold {|c|}}.should raise_error(Sequel::DatabaseConnectionError)
+  end
+  
+  specify "should raise a DatabaseConnectionError if the connection proc returns nil" do
+    db = Sequel::Database.new(:single_threaded => true, :servers=>{}){nil}
+    proc {db.pool.hold {|c|}}.should raise_error(Sequel::DatabaseConnectionError)
   end
 end
 
@@ -820,7 +1118,7 @@ context "A database" do
   end
   
   specify "should have single_threaded? respond to true if in single threaded mode" do
-    db = Sequel::Database.new(:single_threaded => true)
+    db = Sequel::Database.new(:single_threaded => true){1234}
     db.should be_single_threaded
     
     db = Sequel::Database.new(:max_options => 1)
@@ -831,10 +1129,10 @@ context "A database" do
     
     Sequel::Database.single_threaded = true
     
-    db = Sequel::Database.new
+    db = Sequel::Database.new{123}
     db.should be_single_threaded
     
-    db = Sequel::Database.new(:max_options => 4)
+    db = Sequel::Database.new(:max_options => 4){123}
     db.should be_single_threaded
   end
   
@@ -877,12 +1175,16 @@ context "Database#fetch" do
     @db.fetch('select * from xyz where x = ? and y = ?', 15, 'abc') {|r| sql = r[:sql]}
     sql.should == "select * from xyz where x = 15 and y = 'abc'"
     
-    # and Aman Gupta's example
-    @db.fetch('select name from table where name = ? or id in ?',
-    'aman', [3,4,7]) {|r| sql = r[:sql]}
+    @db.fetch('select name from table where name = ? or id in ?', 'aman', [3,4,7]) {|r| sql = r[:sql]}
     sql.should == "select name from table where name = 'aman' or id in (3, 4, 7)"
   end
   
+  specify "should format the given sql with named arguments" do
+    sql = nil
+    @db.fetch('select * from xyz where x = :x and y = :y', :x=>15, :y=>'abc') {|r| sql = r[:sql]}
+    sql.should == "select * from xyz where x = 15 and y = 'abc'"
+  end
+  
   specify "should return the dataset if no block is given" do
     @db.fetch('select * from xyz').should be_a_kind_of(Sequel::Dataset)
     
@@ -915,7 +1217,7 @@ context "Database#[]" do
     ds.opts[:from].should == [:items]
   end
   
-  specify "should return an enumerator when a string is given" do
+  specify "should return a dataset when a string is given" do
     c = Class.new(Sequel::Dataset) do
       def fetch_rows(sql); yield({:sql => sql}); end
     end
@@ -1054,22 +1356,22 @@ end
 context "Database#server_opts" do
   specify "should return the general opts if no :servers option is used" do
     opts = {:host=>1, :database=>2}
-    MockDatabase.new(opts).send(:server_opts, :server1).should == {:host=>1, :database=>2}
+    MockDatabase.new(opts).send(:server_opts, :server1)[:host].should == 1
   end
   
   specify "should return the general opts if entry for the server is present in the :servers option" do
     opts = {:host=>1, :database=>2, :servers=>{}}
-    MockDatabase.new(opts).send(:server_opts, :server1).should == {:host=>1, :database=>2}
+    MockDatabase.new(opts).send(:server_opts, :server1)[:host].should == 1
   end
   
   specify "should return the general opts merged with the specific opts if given as a hash" do
     opts = {:host=>1, :database=>2, :servers=>{:server1=>{:host=>3}}}
-    MockDatabase.new(opts).send(:server_opts, :server1).should == {:host=>3, :database=>2}
+    MockDatabase.new(opts).send(:server_opts, :server1)[:host].should == 3
   end
   
   specify "should return the sgeneral opts merged with the specific opts if given as a proc" do
     opts = {:host=>1, :database=>2, :servers=>{:server1=>proc{|db| {:host=>4}}}}
-    MockDatabase.new(opts).send(:server_opts, :server1).should == {:host=>4, :database=>2}
+    MockDatabase.new(opts).send(:server_opts, :server1)[:host].should == 4
   end
   
   specify "should raise an error if the specific opts is not a proc or hash" do
@@ -1078,6 +1380,140 @@ context "Database#server_opts" do
   end
 end
 
+context "Database#add_servers" do
+  before do
+    @db = MockDatabase.new(:host=>1, :database=>2, :servers=>{:server1=>{:host=>3}})
+    def @db.connect(server)
+      server_opts(server)
+    end
+    def @db.disconnect_connection(c)
+    end
+  end
+
+  specify "should add new servers to the connection pool" do
+    @db.synchronize{|c| c[:host].should == 1}
+    @db.synchronize(:server1){|c| c[:host].should == 3}
+    @db.synchronize(:server2){|c| c[:host].should == 1}
+
+    @db.add_servers(:server2=>{:host=>6})
+    @db.synchronize{|c| c[:host].should == 1}
+    @db.synchronize(:server1){|c| c[:host].should == 3}
+    @db.synchronize(:server2){|c| c[:host].should == 6}
+
+    @db.disconnect
+    @db.synchronize{|c| c[:host].should == 1}
+    @db.synchronize(:server1){|c| c[:host].should == 3}
+    @db.synchronize(:server2){|c| c[:host].should == 6}
+  end
+
+  specify "should replace options for future connections to existing servers" do
+    @db.synchronize{|c| c[:host].should == 1}
+    @db.synchronize(:server1){|c| c[:host].should == 3}
+    @db.synchronize(:server2){|c| c[:host].should == 1}
+
+    @db.add_servers(:default=>proc{{:host=>4}}, :server1=>{:host=>8})
+    @db.synchronize{|c| c[:host].should == 1}
+    @db.synchronize(:server1){|c| c[:host].should == 3}
+    @db.synchronize(:server2){|c| c[:host].should == 1}
+
+    @db.disconnect
+    @db.synchronize{|c| c[:host].should == 4}
+    @db.synchronize(:server1){|c| c[:host].should == 8}
+    @db.synchronize(:server2){|c| c[:host].should == 4}
+  end
+end
+
+context "Database#remove_servers" do
+  before do
+    @db = MockDatabase.new(:host=>1, :database=>2, :servers=>{:server1=>{:host=>3}, :server2=>{:host=>4}})
+    def @db.connect(server)
+      server_opts(server)
+    end
+    def @db.disconnect_connection(c)
+    end
+  end
+
+  specify "should remove servers from the connection pool" do
+    @db.synchronize{|c| c[:host].should == 1}
+    @db.synchronize(:server1){|c| c[:host].should == 3}
+    @db.synchronize(:server2){|c| c[:host].should == 4}
+
+    @db.remove_servers(:server1, :server2)
+    @db.synchronize{|c| c[:host].should == 1}
+    @db.synchronize(:server1){|c| c[:host].should == 1}
+    @db.synchronize(:server2){|c| c[:host].should == 1}
+  end
+  
+  specify "should accept arrays of symbols" do
+    @db.remove_servers([:server1, :server2])
+    @db.synchronize{|c| c[:host].should == 1}
+    @db.synchronize(:server1){|c| c[:host].should == 1}
+    @db.synchronize(:server2){|c| c[:host].should == 1}
+  end
+
+  specify "should allow removal while connections are still open" do
+    @db.synchronize do |c1|
+      c1[:host].should == 1
+      @db.synchronize(:server1) do |c2|
+        c2[:host].should == 3
+        @db.synchronize(:server2) do |c3|
+          c3[:host].should == 4
+          @db.remove_servers(:server1, :server2)
+            @db.synchronize(:server1) do |c4|
+              c4.should_not == c2
+              c4.should == c1
+              c4[:host].should == 1
+              @db.synchronize(:server2) do |c5|
+                c5.should_not == c3
+                c5.should == c1
+                c5[:host].should == 1
+              end
+            end
+          c3[:host].should == 4
+         end
+        c2[:host].should == 3
+      end
+      c1[:host].should == 1
+    end
+  end
+end
+
+context "Database#each_server" do
+  before do
+    @db = Sequel.connect(:adapter=>:mock, :host=>1, :database=>2, :servers=>{:server1=>{:host=>3}, :server2=>{:host=>4}})
+    def @db.connect(server)
+      server_opts(server)
+    end
+    def @db.disconnect_connection(c)
+    end
+  end
+
+  specify "should yield a separate database object for each server" do
+    hosts = []
+    @db.each_server do |db|
+      db.should be_a_kind_of(Sequel::Database)
+      db.should_not == @db
+      db.opts[:database].should == 2
+      hosts << db.opts[:host]
+    end
+    hosts.sort.should == [1, 3, 4]
+  end
+
+  specify "should disconnect and remove entry from Sequel::DATABASES after use" do
+    dbs = []
+    dcs = []
+    @db.each_server do |db|
+      dbs << db
+      Sequel::DATABASES.should include(db)
+      db.meta_def(:disconnect){dcs << db}
+    end
+    dbs.each do |db|
+      Sequel::DATABASES.should_not include(db)
+    end
+    dbs.should == dcs
+  end
+end
+  
 context "Database#raise_error" do
   specify "should reraise if the exception class is not in opts[:classes]" do
     e = Class.new(StandardError)
@@ -1091,12 +1527,17 @@ context "Database#raise_error" do
   specify "should convert the exception to a DatabaseError if opts[:classes] if not present" do
     proc{MockDatabase.new.send(:raise_error, Interrupt.new(''))}.should raise_error(Sequel::DatabaseError)
   end
+  
+  specify "should convert the exception to a DatabaseDisconnectError if opts[:disconnect] is true" do
+    proc{MockDatabase.new.send(:raise_error, Interrupt.new(''), :disconnect=>true)}.should raise_error(Sequel::DatabaseDisconnectError)
+  end
 end
 
 context "Database#typecast_value" do
   before do
     @db = Sequel::Database.new
   end
+
   specify "should raise an InvalidValue when given an invalid value" do
     proc{@db.typecast_value(:integer, "13a")}.should raise_error(Sequel::InvalidValue)
     proc{@db.typecast_value(:float, "4.e2")}.should raise_error(Sequel::InvalidValue)
@@ -1106,6 +1547,24 @@ context "Database#typecast_value" do
     proc{@db.typecast_value(:time, Date.new)}.should raise_error(Sequel::InvalidValue)
     proc{@db.typecast_value(:datetime, 4)}.should raise_error(Sequel::InvalidValue)
   end
+
+  specify "should have an underlying exception class available at wrapped_exception" do
+    begin
+      @db.typecast_value(:date, 'a')
+      true.should == false
+    rescue Sequel::InvalidValue => e
+      e.wrapped_exception.should be_a_kind_of(ArgumentError)
+    end
+  end
+
+  specify "should include underlying exception class in #inspect" do
+    begin
+      @db.typecast_value(:date, 'a')
+      true.should == false
+    rescue Sequel::InvalidValue => e
+      e.inspect.should =~ /\A#<Sequel::InvalidValue: ArgumentError: .*>\z/
+    end
+  end
 end
 
 context "Database#blank_object?" do
@@ -1140,3 +1599,113 @@ context "Database#schema_autoincrementing_primary_key?" do
     m.call(:primary_key=>false).should == false
   end
 end
+
+context "Database#supports_savepoints?" do
+  specify "should be false by default" do
+    Sequel::Database.new.supports_savepoints?.should == false
+  end
+end
+
+context "Database#supports_prepared_transactions?" do
+  specify "should be false by default" do
+    Sequel::Database.new.supports_prepared_transactions?.should == false
+  end
+end
+
+context "Database#supports_transaction_isolation_levels?" do
+  specify "should be false by default" do
+    Sequel::Database.new.supports_transaction_isolation_levels?.should == false
+  end
+end
+
+context "Database#input_identifier_meth" do
+  specify "should be the input_identifer method of a default dataset for this database" do
+    db = Sequel::Database.new
+    db.send(:input_identifier_meth).call(:a).should == 'a'
+    db.identifier_input_method = :upcase
+    db.send(:input_identifier_meth).call(:a).should == 'A'
+  end
+end
+
+context "Database#output_identifier_meth" do
+  specify "should be the output_identifer method of a default dataset for this database" do
+    db = Sequel::Database.new
+    db.send(:output_identifier_meth).call('A').should == :A
+    db.identifier_output_method = :downcase
+    db.send(:output_identifier_meth).call('A').should == :a
+  end
+end
+
+context "Database#metadata_dataset" do
+  specify "should be a dataset with the default settings for identifier_input_method and identifier_output_method" do
+    ds = Sequel::Database.new.send(:metadata_dataset)
+    ds.literal(:a).should == 'A'
+    ds.send(:output_identifier, 'A').should == :a
+  end
+end
+
+context "Database#column_schema_to_ruby_default" do
+  specify "should handle converting many default formats" do
+    db = Sequel::Database.new
+    m = db.method(:column_schema_to_ruby_default)
+    p = lambda{|d,t| m.call(d,t)}
+    p[nil, :integer].should == nil
+    p['1', :integer].should == 1
+    p['-1', :integer].should == -1
+    p['1.0', :float].should == 1.0
+    p['-1.0', :float].should == -1.0
+    p['1.0', :decimal].should == BigDecimal.new('1.0')
+    p['-1.0', :decimal].should == BigDecimal.new('-1.0')
+    p['1', :boolean].should == true
+    p['0', :boolean].should == false
+    p['true', :boolean].should == true
+    p['false', :boolean].should == false
+    p["'t'", :boolean].should == true
+    p["'f'", :boolean].should == false
+    p["'a'", :string].should == 'a'
+    p["'a'", :blob].should == 'a'.to_sequel_blob
+    p["'a'", :blob].should be_a_kind_of(Sequel::SQL::Blob)
+    p["''", :string].should == ''
+    p["'\\a''b'", :string].should == "\\a'b"
+    p["'NULL'", :string].should == "NULL"
+    p["'2009-10-29'", :date].should == Date.new(2009,10,29)
+    p["CURRENT_TIMESTAMP", :date].should == nil
+    p["today()", :date].should == nil
+    p["'2009-10-29T10:20:30-07:00'", :datetime].should == DateTime.parse('2009-10-29T10:20:30-07:00')
+    p["'2009-10-29 10:20:30'", :datetime].should == DateTime.parse('2009-10-29 10:20:30')
+    p["'10:20:30'", :time].should == Time.parse('10:20:30')
+    p["NaN", :float].should == nil
+
+    db.meta_def(:database_type){:postgres}
+    p["''::text", :string].should == ""
+    p["'\\a''b'::character varying", :string].should == "\\a'b"
+    p["'a'::bpchar", :string].should == "a"
+    p["(-1)", :integer].should == -1
+    p["(-1.0)", :float].should == -1.0
+    p['(-1.0)', :decimal].should == BigDecimal.new('-1.0')
+    p["'a'::bytea", :blob].should == 'a'.to_sequel_blob
+    p["'a'::bytea", :blob].should be_a_kind_of(Sequel::SQL::Blob)
+    p["'2009-10-29'::date", :date].should == Date.new(2009,10,29)
+    p["'2009-10-29 10:20:30.241343'::timestamp without time zone", :datetime].should == DateTime.parse('2009-10-29 10:20:30.241343')
+    p["'10:20:30'::time without time zone", :time].should == Time.parse('10:20:30')
+
+    db.meta_def(:database_type){:mysql}
+    p["\\a'b", :string].should == "\\a'b"
+    p["a", :string].should == "a"
+    p["NULL", :string].should == "NULL"
+    p["-1", :float].should == -1.0
+    p['-1', :decimal].should == BigDecimal.new('-1.0')
+    p["2009-10-29", :date].should == Date.new(2009,10,29)
+    p["2009-10-29 10:20:30", :datetime].should == DateTime.parse('2009-10-29 10:20:30')
+    p["10:20:30", :time].should == Time.parse('10:20:30')
+    p["CURRENT_DATE", :date].should == nil
+    p["CURRENT_TIMESTAMP", :datetime].should == nil
+    p["a", :enum].should == "a"
+    
+    db.meta_def(:database_type){:mssql}
+    p["(N'a')", :string].should == "a"
+    p["((-12))", :integer].should == -12
+    p["((12.1))", :float].should == 12.1
+    p["((-12.1))", :decimal].should == BigDecimal.new('-12.1')
+  end
+end
