@@ -1,4 +1,4 @@
-require File.join(File.dirname(__FILE__), 'spec_helper.rb')
+require File.join(File.dirname(File.expand_path(__FILE__)), 'spec_helper.rb')
 
 describe "Simple Dataset operations" do
   before do
@@ -38,11 +38,18 @@ describe "Simple Dataset operations" do
     @ds.filter(1=>1).delete.should == 1
     @ds.count.should == 0
   end
-
+  
   specify "should update correctly" do
     @ds.update(:number=>:number+1).should == 1
     @ds.all.should == [{:id=>1, :number=>11}]
   end
+  
+  cspecify "should have update return the number of matched rows", [:mysql, :mysql], [:do, :mysql], [:ado] do
+    @ds.update(:number=>:number).should == 1
+    @ds.filter(:id=>1).update(:number=>:number).should == 1
+    @ds.filter(:id=>2).update(:number=>:number).should == 0
+    @ds.all.should == [{:id=>1, :number=>10}]
+  end
 
   specify "should fetch all results correctly" do
     @ds.all.should == [{:id=>1, :number=>10}]
@@ -52,6 +59,10 @@ describe "Simple Dataset operations" do
     @ds.first.should == {:id=>1, :number=>10}
   end
   
+  specify "should have distinct work with limit" do
+    @ds.limit(1).distinct.all.should == [{:id=>1, :number=>10}]
+  end
+  
   specify "should fetch correctly with a limit" do
     @ds.order(:id).limit(2).all.should == [{:id=>1, :number=>10}]
     @ds.insert(:number=>20)
@@ -266,6 +277,10 @@ describe "Simple Dataset operations in transactions" do
       @ds.order(:id).all.should == [{:id=>1, :number=>20}]
     end
   end
+  
+  specify "should support for_update" do
+    INTEGRATION_DB.transaction{@ds.for_update.all.should == []}
+  end
 end
 
 describe "Dataset UNION, EXCEPT, and INTERSECT" do
@@ -536,23 +551,23 @@ describe "Sequel::Dataset convenience methods" do
   end
   
   it "#group_and_count should return a grouping by count" do
-    @ds.group_and_count(:a).all.should == []
+    @ds.group_and_count(:a).order(:count).all.should == []
     @ds.insert(20, 10)
-    @ds.group_and_count(:a).all.each{|h| h[:count] = h[:count].to_i}.should == [{:a=>20, :count=>1}]
+    @ds.group_and_count(:a).order(:count).all.each{|h| h[:count] = h[:count].to_i}.should == [{:a=>20, :count=>1}]
     @ds.insert(20, 30)
-    @ds.group_and_count(:a).all.each{|h| h[:count] = h[:count].to_i}.should == [{:a=>20, :count=>2}]
+    @ds.group_and_count(:a).order(:count).all.each{|h| h[:count] = h[:count].to_i}.should == [{:a=>20, :count=>2}]
     @ds.insert(30, 30)
-    @ds.group_and_count(:a).all.each{|h| h[:count] = h[:count].to_i}.should == [{:a=>30, :count=>1}, {:a=>20, :count=>2}]
+    @ds.group_and_count(:a).order(:count).all.each{|h| h[:count] = h[:count].to_i}.should == [{:a=>30, :count=>1}, {:a=>20, :count=>2}]
   end
   
   it "#group_and_count should support column aliases" do
-    @ds.group_and_count(:a___c).all.should == []
+    @ds.group_and_count(:a___c).order(:count).all.should == []
     @ds.insert(20, 10)
-    @ds.group_and_count(:a___c).all.each{|h| h[:count] = h[:count].to_i}.should == [{:c=>20, :count=>1}]
+    @ds.group_and_count(:a___c).order(:count).all.each{|h| h[:count] = h[:count].to_i}.should == [{:c=>20, :count=>1}]
     @ds.insert(20, 30)
-    @ds.group_and_count(:a___c).all.each{|h| h[:count] = h[:count].to_i}.should == [{:c=>20, :count=>2}]
+    @ds.group_and_count(:a___c).order(:count).all.each{|h| h[:count] = h[:count].to_i}.should == [{:c=>20, :count=>2}]
     @ds.insert(30, 30)
-    @ds.group_and_count(:a___c).all.each{|h| h[:count] = h[:count].to_i}.should == [{:c=>30, :count=>1}, {:c=>20, :count=>2}]
+    @ds.group_and_count(:a___c).order(:count).all.each{|h| h[:count] = h[:count].to_i}.should == [{:c=>30, :count=>1}, {:c=>20, :count=>2}]
   end
   
   cspecify "#range should return the range between the maximum and minimum values", :sqlite do
@@ -596,7 +611,7 @@ describe "Sequel::Dataset main SQL methods" do
     @ds.filter("b < ?", 15).invert.all.should == [{:a=>20, :b=>30}]
   end
   
-  it "#and and #or should work with placeholder strings" do
+  it "#and and #or should work correctly" do
     @ds.insert(20, 30)
     @ds.filter(:a=>20).and(:b=>30).all.should == [{:a=>20, :b=>30}]
     @ds.filter(:a=>20).and(:b=>15).all.should == []
@@ -641,32 +656,47 @@ describe "Sequel::Dataset DSL support" do
     @ds.get{a / b}.to_i.should == 2
   end
   
-  cspecify "should work with standard bitwise mathematical operators", :mssql, :h2 do
-    @ds.insert(24, 2)
-    @ds.get{a.sql_number << b}.to_i.should == 96
-    @ds.get{a.sql_number >> b}.to_i.should == 6
+  specify "should work with bitwise shift operators" do
+    @ds.insert(3, 2)
+    @ds.get{a.sql_number << b}.to_i.should == 12
+    @ds.get{a.sql_number >> b}.to_i.should == 0
     @ds.delete
+    @ds.insert(3, 1)
+    @ds.get{a.sql_number << b}.to_i.should == 6
+    @ds.get{a.sql_number >> b}.to_i.should == 1
+  end
+
+  specify "should work with bitwise AND and OR operators" do
     @ds.insert(3, 5)
     @ds.get{a.sql_number | b}.to_i.should == 7
     @ds.get{a.sql_number & b}.to_i.should == 1
   end
   
-  cspecify "should work with the bitwise compliment operator", :mysql, :h2 do
+  cspecify "should work with the bitwise compliment operator", :h2 do
+    @ds.insert(-3, 3)
+    @ds.get{~a.sql_number}.to_i.should == 2
+    @ds.get{~b.sql_number}.to_i.should == -4
+  end
+  
+  cspecify "should work with the bitwise xor operator", :sqlite do
     @ds.insert(3, 5)
-    @ds.get{~a.sql_number}.to_i.should == -4
+    @ds.get{a.sql_number ^ b}.to_i.should == 6
   end
   
-  cspecify "should work with inequality operators", :mssql do
+  specify "should work with inequality operators" do
+    @ds.insert(10, 11)
+    @ds.insert(11, 11)
+    @ds.insert(20, 19)
     @ds.insert(20, 20)
-    ['0', 0, false].should include(@ds.get{a > b})
-    ['0', 0, false].should include(@ds.get{a < b})
-    ['1', 1, true].should include(@ds.get{a <= b})
-    ['1', 1, true].should include(@ds.get{a >= b})
+    @ds.filter{a > b}.select_order_map(:a).should == [20]
+    @ds.filter{a >= b}.select_order_map(:a).should == [11, 20, 20]
+    @ds.filter{a < b}.select_order_map(:a).should == [10]
+    @ds.filter{a <= b}.select_order_map(:a).should == [10, 11, 20]
   end
   
-  cspecify "should work with casting and string concatentation", :mssql do
+  specify "should work with casting and string concatentation" do
     @ds.insert(20, 20)
-    @ds.get{a.cast_string + b}.should == '2020'
+    @ds.get{a.cast_string + b.cast_string}.should == '2020'
   end
   
   it "should work with ordering" do
@@ -715,6 +745,14 @@ describe "Sequel::Dataset DSL support" do
     @ds.exclude(:a=>nil).all.should == [{:a=>20, :b=>nil}]
   end
   
+  it "should work with arrays as hash values" do
+    @ds.insert(20, 10)
+    @ds.filter(:a=>[10]).all.should == []
+    @ds.filter(:a=>[20, 10]).all.should == [{:a=>20, :b=>10}]
+    @ds.exclude(:a=>[10]).all.should == [{:a=>20, :b=>10}]
+    @ds.exclude(:a=>[20, 10]).all.should == []
+  end
+  
   it "should work with ranges as hash values" do
     @ds.insert(20, 10)
     @ds.filter(:a=>(10..30)).all.should == [{:a=>20, :b=>10}]
@@ -733,7 +771,7 @@ describe "Sequel::Dataset DSL support" do
     @ds.filter({15=>20}.case(0, :a) > 0).all.should == []
   end
   
-  it "should work multiple value arrays" do
+  it "should work with multiple value arrays" do
     @ds.insert(20, 10)
     @ds.quote_identifiers = false
     @ds.filter([:a, :b]=>[[20, 10]].sql_array).all.should == [{:a=>20, :b=>10}]
@@ -746,6 +784,41 @@ describe "Sequel::Dataset DSL support" do
     @ds.exclude([:a, :b]=>[[20, 10], [1, 2]].sql_array).all.should == []
     @ds.exclude([:a, :b]=>[[10, 10], [20, 20]].sql_array).all.should == [{:a=>20, :b=>10}]
   end
+
+  it "should work with IN/NOT in with datasets" do
+    @ds.insert(20, 10)
+    ds = @ds.unordered
+    @ds.quote_identifiers = false
+
+    @ds.filter(:a=>ds.select(:a)).all.should == [{:a=>20, :b=>10}]
+    @ds.filter(:a=>ds.select(:a).where(:a=>15)).all.should == []
+    @ds.exclude(:a=>ds.select(:a)).all.should == []
+    @ds.exclude(:a=>ds.select(:a).where(:a=>15)).all.should == [{:a=>20, :b=>10}]
+
+    @ds.filter([:a, :b]=>ds.select(:a, :b)).all.should == [{:a=>20, :b=>10}]
+    @ds.filter([:a, :b]=>ds.select(:b, :a)).all.should == []
+    @ds.exclude([:a, :b]=>ds.select(:a, :b)).all.should == []
+    @ds.exclude([:a, :b]=>ds.select(:b, :a)).all.should == [{:a=>20, :b=>10}]
+
+    @ds.filter([:a, :b]=>ds.select(:a, :b).where(:a=>15)).all.should == []
+    @ds.exclude([:a, :b]=>ds.select(:a, :b).where(:a=>15)).all.should == [{:a=>20, :b=>10}]
+  end
+
+  specify "should work empty arrays" do
+    @ds.insert(20, 10)
+    @ds.filter(:a=>[]).all.should == []
+    @ds.exclude(:a=>[]).all.should == [{:a=>20, :b=>10}]
+    @ds.filter([:a, :b]=>[]).all.should == []
+    @ds.exclude([:a, :b]=>[]).all.should == [{:a=>20, :b=>10}]
+  end
+  
+  specify "should work empty arrays with nulls" do
+    @ds.insert(nil, nil)
+    @ds.filter(:a=>[]).all.should == []
+    @ds.exclude(:a=>[]).all.should == [{:a=>nil, :b=>nil}]
+    @ds.filter([:a, :b]=>[]).all.should == []
+    @ds.exclude([:a, :b]=>[]).all.should == [{:a=>nil, :b=>nil}]
+  end
   
   it "should work multiple conditions" do
     @ds.insert(20, 10)
@@ -870,3 +943,40 @@ describe "Dataset defaults and overrides" do
     @ds.all.should == [{:a=>10}, {:a=>10}]
   end
 end
+
+if INTEGRATION_DB.dataset.supports_modifying_joins?
+  describe "Modifying joined datasets" do
+    before do
+      @db = INTEGRATION_DB
+      @db.create_table!(:a){Integer :a; Integer :d}
+      @db.create_table!(:b){Integer :b; Integer :e}
+      @db.create_table!(:c){Integer :c; Integer :f}
+      @ds = @db.from(:a, :b).join(:c, :c=>:e.identifier).where(:d=>:b, :f=>6)
+      @db[:a].insert(1, 2)
+      @db[:a].insert(3, 4)
+      @db[:b].insert(2, 5)
+      @db[:c].insert(5, 6)
+      @db[:b].insert(4, 7)
+      @db[:c].insert(7, 8)
+    end
+    after do
+      @db.drop_table(:a, :b, :c)
+    end
+    
+    it "#update should allow updating joined datasets" do
+      @ds.update(:a=>10)
+      @ds.all.should == [{:c=>5, :b=>2, :a=>10, :d=>2, :e=>5, :f=>6}]
+      @db[:a].order(:a).all.should == [{:a=>3, :d=>4}, {:a=>10, :d=>2}]
+      @db[:b].order(:b).all.should == [{:b=>2, :e=>5}, {:b=>4, :e=>7}]
+      @db[:c].order(:c).all.should == [{:c=>5, :f=>6}, {:c=>7, :f=>8}]
+    end
+    
+    it "#delete should allow deleting from joined datasets" do
+      @ds.delete
+      @ds.all.should == []
+      @db[:a].order(:a).all.should == [{:a=>3, :d=>4}]
+      @db[:b].order(:b).all.should == [{:b=>2, :e=>5}, {:b=>4, :e=>7}]
+      @db[:c].order(:c).all.should == [{:c=>5, :f=>6}, {:c=>7, :f=>8}]
+    end
+  end
+end
